@@ -1,6 +1,6 @@
 /*
  * cups - Python bindings for CUPS
- * Copyright (C) 2002, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012  Red Hat, Inc
+ * Copyright (C) 2002, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013  Red Hat, Inc
  * Author: Tim Waugh <twaugh@redhat.com>
  *
  * This program is free software; you can redistribute it and/or modify
@@ -189,7 +189,7 @@ do_password_callback (const char *prompt)
     password = NULL;
   else
   {
-    pwval = PyString_AsString (result);
+    pwval = PyBytes_AsString (result);
     password = strdup (pwval);
   }
 
@@ -396,9 +396,13 @@ cups_modelSort (PyObject *self, PyObject *args)
     PyErr_SetString (PyExc_RuntimeError, "Insufficient memory");
     return NULL;
   }
-
+#if PY_MAJOR_VERSION >= 3
+  PyUnicode_AsWideChar (a, wca, size_a);
+  PyUnicode_AsWideChar (b, wcb, size_b);
+#else
   PyUnicode_AsWideChar ((PyUnicodeObject *) a, wca, size_a);
   PyUnicode_AsWideChar ((PyUnicodeObject *) b, wcb, size_b);
+#endif
   Py_DECREF (a);
   Py_DECREF (b);
   return Py_BuildValue ("i", do_model_compare (wca, wcb));
@@ -407,24 +411,34 @@ cups_modelSort (PyObject *self, PyObject *args)
 static PyObject *
 cups_setUser (PyObject *self, PyObject *args)
 {
-  const char *user;
+  PyObject *userobj;
+  char *user;
 
-  if (!PyArg_ParseTuple (args, "s", &user))
+  if (!PyArg_ParseTuple (args, "O", &userobj))
+    return NULL;
+
+  if (UTF8_from_PyObj (&user, userobj) == NULL)
     return NULL;
 
   cupsSetUser (user);
+  free (user);
   Py_RETURN_NONE;
 }
 
 static PyObject *
 cups_setServer (PyObject *self, PyObject *args)
 {
-  const char *server;
+  PyObject *serverobj;
+  char *server;
 
-  if (!PyArg_ParseTuple (args, "s", &server))
+  if (!PyArg_ParseTuple (args, "O", &serverobj))
+    return NULL;
+
+  if (UTF8_from_PyObj (&server, serverobj) == NULL)
     return NULL;
 
   cupsSetServer (server);
+  free (server);
   Py_RETURN_NONE;
 }
 
@@ -454,13 +468,13 @@ cups_setEncryption (PyObject *self, PyObject *args)
 static PyObject *
 cups_getUser (PyObject *self)
 {
-  return PyString_FromString (cupsUser ());
+  return PyUnicode_FromString (cupsUser ());
 }
 
 static PyObject *
 cups_getServer (PyObject *self)
 {
-  return PyString_FromString (cupsServer ());
+  return PyUnicode_FromString (cupsServer ());
 }
 
 static PyObject *
@@ -565,6 +579,159 @@ cups_ppdSetConformance (PyObject *self, PyObject *args)
   Py_RETURN_NONE;
 }
 
+#ifdef HAVE_CUPS_1_6
+static PyObject *
+cups_enumDests (PyObject *self, PyObject *args, PyObject *kwds)
+{
+  PyObject *cb;
+  int flags = 0;
+  int msec = -1;
+  int type = 0;
+  int mask = 0;
+  PyObject *user_data = NULL;
+  CallbackContext context;
+  int ret;
+  static char *kwlist[] = { "cb",
+			    "flags",
+			    "msec",
+			    "type",
+			    "mask",
+			    "user_data",
+			    NULL };
+
+  if (!PyArg_ParseTupleAndKeywords (args, kwds, "O|iiiiO", kwlist,
+				    &cb,
+				    &flags,
+				    &msec,
+				    &type,
+				    &mask,
+				    &user_data))
+    return NULL;
+
+  if (!PyCallable_Check (cb)) {
+    PyErr_SetString (PyExc_TypeError, "cb must be callable");
+    return NULL;
+  }
+
+  if (!user_data)
+    user_data = Py_None;
+
+  Py_XINCREF (cb);
+  Py_XINCREF (user_data);
+  context.cb = cb;
+  context.user_data = user_data;
+  ret = cupsEnumDests (flags,
+		       msec,
+		       NULL,
+		       type,
+		       mask,
+		       cups_dest_cb,
+		       &context);
+  Py_XDECREF (cb);
+  Py_XDECREF (user_data);
+
+  if (!ret) {
+    PyErr_SetString (PyExc_RuntimeError, "cupsEnumDests failed");
+    return NULL;
+  }
+
+  Py_RETURN_NONE;
+}
+#endif /* HAVE_CUPS_1_6 */
+
+#ifdef HAVE_CUPS_1_6
+static PyObject *
+cups_connectDest (PyObject *self, PyObject *args, PyObject *kwds)
+{
+  PyObject *destobj;
+  PyObject *cb;
+  int flags = 0;
+  int msec = -1;
+  PyObject *user_data = NULL;
+  CallbackContext context;
+  char resource[HTTP_MAX_URI];
+  http_t *conn;
+  Connection *connobj;
+  Dest *dest_o;
+  cups_dest_t dest;
+  PyObject *ret;
+  static char *kwlist[] = { "dest",
+			    "cb",
+			    "flags",
+			    "msec",
+			    "user_data",
+			    NULL };
+
+  if (!PyArg_ParseTupleAndKeywords (args, kwds, "OO|iiO", kwlist,
+				    &destobj,
+				    &cb,
+				    &flags,
+				    &msec,
+				    &user_data))
+    return NULL;
+
+  if (Py_TYPE(destobj) != &cups_DestType) {
+    PyErr_SetString (PyExc_TypeError, "dest must be Dest object");
+    return NULL;
+  }
+
+  if (!PyCallable_Check (cb)) {
+    PyErr_SetString (PyExc_TypeError, "cb must be callable");
+    return NULL;
+  }
+
+  if (!user_data)
+    user_data = Py_None;
+
+  Py_XINCREF (cb);
+  Py_XINCREF (user_data);
+  context.cb = cb;
+  context.user_data = user_data;
+  resource[0] = '\0';
+
+  dest_o = (Dest *) destobj;
+  dest.is_default = dest_o->is_default;
+  dest.name = dest_o->destname;
+  dest.instance = dest_o->instance;
+  dest.num_options = dest_o->num_options;
+  dest.options = malloc (dest_o->num_options * sizeof (dest.options[0]));
+  int i;
+  for (i = 0; i < dest_o->num_options; i++) {
+    dest.options[i].name = dest_o->name[i];
+    dest.options[i].value = dest_o->value[i];
+  }
+
+  conn = cupsConnectDest (&dest,
+			  flags,
+			  msec,
+			  NULL,
+			  resource,
+			  sizeof (resource),
+			  cups_dest_cb,
+			  &context);
+  Py_XDECREF (cb);
+  Py_XDECREF (user_data);
+  free (dest.options);
+
+  if (!conn) {
+    set_ipp_error (cupsLastError (), cupsLastErrorString ());
+    return NULL;
+  }
+
+  PyObject *largs = Py_BuildValue ("()");
+  PyObject *lkwlist = Py_BuildValue ("{}");
+  connobj = (Connection *) PyType_GenericNew (&cups_ConnectionType,
+					      largs, lkwlist);
+  Py_DECREF (largs);
+  Py_DECREF (lkwlist);
+  connobj->host = strdup ("");
+  connobj->http = conn;
+
+  ret = Py_BuildValue ("(Os)", (PyObject *) connobj, resource);
+  return ret;
+}
+#endif /* HAVE_CUPS_1_6 */
+
 static PyObject *
 cups_require (PyObject *self, PyObject *args)
 {
@@ -607,7 +774,19 @@ fail:
   return NULL;
 }
 
-static PyMethodDef CupsMethods[] = {
+struct module_state {
+    PyObject *error;
+};
+
+#if PY_MAJOR_VERSION >= 3
+#define GETSTATE(m) ((struct module_state*)PyModule_GetState(m))
+#else
+#define GETSTATE(m) (&_state)
+static struct module_state _state;
+#endif
+
+
+static PyMethodDef cups_methods[] = {
   { "modelSort", cups_modelSort, METH_VARARGS,
     "modelSort(s1,s2) -> integer\n\n"
     "Sort two model strings.\n\n"
@@ -687,6 +866,43 @@ static PyMethodDef CupsMethods[] = {
     "@type level: integer\n"
     "@param level: PPD_CONFORM_RELAXED or PPD_CONFORM_STRICT" },
 
+#ifdef HAVE_CUPS_1_6
+  { "enumDests",
+    (PyCFunction) cups_enumDests, METH_VARARGS | METH_KEYWORDS,
+    "enumDests(cb,flags=0,msec=-1,type=0,mask=0,user_data=None) -> None\n\n"
+    "@type cb: callable\n"
+    "@param cb: callback function, given user_data, dest flags, and dest.\n"
+    "Should return 1 to continue enumeration and 0 to cancel.\n"
+    "@type flags: integer\n"
+    "@param flags: enumeration flags\n"
+    "@type msec: integer\n"
+    "@param msec: timeout, or -1 for no timeout\n"
+    "@type type: integer\n"
+    "@param type: bitmask of printer types to return\n"
+    "@type mask: integer\n"
+    "@param mask: bitmask of type bits to examine\n"
+    "@type user_data: object\n"
+    "@param user_data: user data to pass to callback function\n"},
+#endif /* HAVE_CUPS_1_6 */
+
+#ifdef HAVE_CUPS_1_6
+  { "connectDest",
+    (PyCFunction) cups_connectDest, METH_VARARGS | METH_KEYWORDS,
+    "connectDest(dest,cb,flags=0,msec=-1,user_data=None) -> (conn, resource)\n\n"
+    "@type dest: Dest object\n"
+    "@param dest: destination to connect to\n"
+    "@type cb: callable\n"
+    "@param cb: callback function, given user_data, dest flags, and dest.\n"
+    "Should return 1 to continue enumeration and 0 to cancel.\n"
+    "@type flags: integer\n"
+    "@param flags: enumeration flags\n"
+    "@type msec: integer\n"
+    "@param msec: timeout, or -1 for no timeout\n"
+    "@type user_data: object\n"
+    "@param user_data: user data to pass to callback function\n"
+    "@return: a 2-tuple of the Connection object and the HTTP resource.\n"},
+#endif /* HAVE_CUPS_1_6 */
+
   { "require", cups_require, METH_VARARGS,
     "require(version) -> None\n\n"
     "Require pycups version.\n\n"
@@ -697,17 +913,65 @@ static PyMethodDef CupsMethods[] = {
   { NULL, NULL, 0, NULL }
 };
 
+#if PY_MAJOR_VERSION >= 3
+static int cups_traverse(PyObject *m, visitproc visit, void *arg) {
+    Py_VISIT(GETSTATE(m)->error);
+    return 0;
+}
+
+static int cups_clear(PyObject *m) {
+    Py_CLEAR(GETSTATE(m)->error);
+    return 0;
+}
+
+
+static struct PyModuleDef moduledef = {
+        PyModuleDef_HEAD_INIT,
+        "cups",
+        NULL,
+        sizeof(struct module_state),
+        cups_methods,
+        NULL,
+        cups_traverse,
+        cups_clear,
+        NULL
+};
+
+#define INITERROR return NULL
+
+PyObject *
+PyInit_cups(void)
+
+#else
+#define INITERROR return
+
 void
 initcups (void)
+#endif
 {
-  PyObject *m = Py_InitModule ("cups", CupsMethods);
+#if PY_MAJOR_VERSION >= 3
+    PyObject *m = PyModule_Create(&moduledef);
+#else
+    PyObject *m = Py_InitModule("cups", cups_methods);
+#endif
+
+  if (m == NULL)
+    INITERROR;
+  struct module_state *st = GETSTATE(m);
+
+  st->error = PyErr_NewException("cups.Error", NULL, NULL);
+  if (st->error == NULL) {
+    Py_DECREF(m);
+    INITERROR;
+  }
+
   PyObject *d = PyModule_GetDict (m);
   PyObject *obj;
 
   // Connection type
   cups_ConnectionType.tp_new = PyType_GenericNew;
   if (PyType_Ready (&cups_ConnectionType) < 0)
-    return;
+    INITERROR;
 
   PyModule_AddObject (m, "Connection",
 		      (PyObject *)&cups_ConnectionType);
@@ -715,7 +979,7 @@ initcups (void)
   // PPD type
   cups_PPDType.tp_new = PyType_GenericNew;
   if (PyType_Ready (&cups_PPDType) < 0)
-    return;
+    INITERROR;
 
   PyModule_AddObject (m, "PPD",
 		      (PyObject *)&cups_PPDType);
@@ -723,7 +987,7 @@ initcups (void)
   // Option type
   cups_OptionType.tp_new = PyType_GenericNew;
   if (PyType_Ready (&cups_OptionType) < 0)
-    return;
+    INITERROR;
 
   PyModule_AddObject (m, "Option",
 		      (PyObject *)&cups_OptionType);
@@ -731,7 +995,7 @@ initcups (void)
   // Group type
   cups_GroupType.tp_new = PyType_GenericNew;
   if (PyType_Ready (&cups_GroupType) < 0)
-    return;
+    INITERROR;
 
   PyModule_AddObject (m, "Group",
 		      (PyObject *)&cups_GroupType);
@@ -739,7 +1003,7 @@ initcups (void)
   // Constraint type
   cups_ConstraintType.tp_new = PyType_GenericNew;
   if (PyType_Ready (&cups_ConstraintType) < 0)
-    return;
+    INITERROR;
 
   PyModule_AddObject (m, "Constraint",
 		      (PyObject *)&cups_ConstraintType);
@@ -747,7 +1011,7 @@ initcups (void)
   // Attribute type
   cups_AttributeType.tp_new = PyType_GenericNew;
   if (PyType_Ready (&cups_AttributeType) < 0)
-    return;
+    INITERROR;
 
   PyModule_AddObject (m, "Attribute",
 		      (PyObject *)&cups_AttributeType);
@@ -755,7 +1019,7 @@ initcups (void)
   // Dest type
   cups_DestType.tp_new = PyType_GenericNew;
   if (PyType_Ready (&cups_DestType) < 0)
-    return;
+    INITERROR;
 
   PyModule_AddObject (m, "Dest",
 		      (PyObject *)&cups_DestType);
@@ -763,7 +1027,7 @@ initcups (void)
   // IPPRequest type
   cups_IPPRequestType.tp_new = PyType_GenericNew;
   if (PyType_Ready (&cups_IPPRequestType) < 0)
-      return;
+      INITERROR;
 
   PyModule_AddObject (m, "IPPRequest",
 		      (PyObject *)&cups_IPPRequestType);
@@ -771,17 +1035,21 @@ initcups (void)
   // IPPAttribute type
   cups_IPPAttributeType.tp_new = PyType_GenericNew;
   if (PyType_Ready (&cups_IPPAttributeType) < 0)
-      return;
+      INITERROR;
 
   PyModule_AddObject (m, "IPPAttribute",
 		      (PyObject *)&cups_IPPAttributeType);
 
   // Constants
-
+#if PY_MAJOR_VERSION >= 3
+  #define INT_CONSTANT(name)					\
+    PyDict_SetItemString (d, #name, PyLong_FromLong (name))
+#else
 #define INT_CONSTANT(name)					\
   PyDict_SetItemString (d, #name, PyInt_FromLong (name))
+#endif
 #define STR_CONSTANT(name)					\
-  PyDict_SetItemString (d, #name, PyString_FromString (name))
+  PyDict_SetItemString (d, #name, PyUnicode_FromString (name))
 
   // CUPS printer types
   INT_CONSTANT (CUPS_PRINTER_LOCAL);
@@ -998,9 +1266,21 @@ initcups (void)
   STR_CONSTANT (CUPS_SERVER_USER_CANCEL_ANY);
   STR_CONSTANT (CUPS_SERVER_REMOTE_ANY);
 
+#ifdef HAVE_CUPS_1_6
+  // Dest enumeration flags
+  INT_CONSTANT (CUPS_DEST_FLAGS_NONE);
+  INT_CONSTANT (CUPS_DEST_FLAGS_UNCONNECTED);
+  INT_CONSTANT (CUPS_DEST_FLAGS_MORE);
+  INT_CONSTANT (CUPS_DEST_FLAGS_REMOVED);
+  INT_CONSTANT (CUPS_DEST_FLAGS_ERROR);
+  INT_CONSTANT (CUPS_DEST_FLAGS_RESOLVING);
+  INT_CONSTANT (CUPS_DEST_FLAGS_CONNECTING);
+  INT_CONSTANT (CUPS_DEST_FLAGS_CANCELED);
+#endif /* HAVE_CUPS_1_6 */
+
   // Exceptions
   obj = PyDict_New ();
-  PyDict_SetItemString (obj, "__doc__", PyString_FromString(
+  PyDict_SetItemString (obj, "__doc__", PyUnicode_FromString(
     "This exception is raised when an HTTP problem has occurred.  It \n"
     "provides an integer HTTP status code.\n\n"
     "Use it like this::\n"
@@ -1011,12 +1291,12 @@ initcups (void)
   HTTPError = PyErr_NewException ("cups.HTTPError", NULL, obj);
   Py_DECREF (obj);
   if (HTTPError == NULL)
-    return;
+    INITERROR;
   Py_INCREF (HTTPError);
   PyModule_AddObject (m, "HTTPError", HTTPError);
 
   obj = PyDict_New ();
-  PyDict_SetItemString (obj, "__doc__", PyString_FromString(
+  PyDict_SetItemString (obj, "__doc__", PyUnicode_FromString(
     "This exception is raised when an IPP error has occurred.  It \n"
     "provides an integer IPP status code, and a human-readable string \n"
     "describing the error.\n\n"
@@ -1029,9 +1309,13 @@ initcups (void)
   IPPError = PyErr_NewException ("cups.IPPError", NULL, obj);
   Py_DECREF (obj);
   if (IPPError == NULL)
-    return;
+    INITERROR;
   Py_INCREF (IPPError);
   PyModule_AddObject (m, "IPPError", IPPError);
+
+#if PY_MAJOR_VERSION >= 3
+    return m;
+#endif
 }
 
 ///////////////
