@@ -209,19 +209,25 @@ PPD_new (PyTypeObject *type, PyObject *args, PyObject *kwds)
 static int
 PPD_init (PPD *self, PyObject *args, PyObject *kwds)
 {
-  const char *filename;
+  PyObject *filenameobj;
+  char *filename;
 
-  if (!PyArg_ParseTuple (args, "s", &filename))
+  if (!PyArg_ParseTuple (args, "O", &filenameobj))
+    return -1;
+
+  if (UTF8_from_PyObj (&filename, filenameobj) == NULL)
     return -1;
 
   self->file = fopen (filename, "r");
   if (!self->file) {
     PyErr_SetString (PyExc_RuntimeError, "fopen failed");
+    free (filename);
     return -1;
   }
 
   debugprintf ("+ PPD %p %s (fd %d)\n", self, filename, fileno (self->file));
   self->ppd = ppdOpenFile (filename);
+  free (filename);
   if (!self->ppd) {
     fclose (self->file);
     self->file = NULL;
@@ -249,7 +255,7 @@ PPD_dealloc (PPD *self)
   if (self->conv_to)
     iconv_close (*self->conv_to);
 
-  self->ob_type->tp_free ((PyObject *) self);
+  ((PyObject *)self)->ob_type->tp_free ((PyObject *) self);
 }
 
 /////////
@@ -268,25 +274,38 @@ static PyObject *
 PPD_localizeIPPReason (PPD *self, PyObject *args, PyObject *kwds)
 {
   PyObject *ret;
-  const char *reason;
-  const char *scheme = NULL;
+  PyObject *reasonobj;
+  PyObject *schemeobj = NULL;
+  char *reason;
+  char *scheme = NULL;
   char *buffer;
   const size_t bufsize = 1024;
   static char *kwlist[] = { "reason", "scheme", NULL };
 
-  if (!PyArg_ParseTupleAndKeywords (args, kwds, "s|s", kwlist,
-				    &reason, &scheme))
+  if (!PyArg_ParseTupleAndKeywords (args, kwds, "O|O", kwlist,
+				    &reasonobj, &schemeobj))
     return NULL;
 
+  if (UTF8_from_PyObj (&reason, reasonobj) == NULL)
+    return NULL;
+
+  if (schemeobj)
+    if (UTF8_from_PyObj (&scheme, schemeobj) == NULL) {
+      free (reason);
+      return NULL;
+    }
+
   buffer = malloc (bufsize);
   if (ppdLocalizeIPPReason (self->ppd, reason, scheme, buffer, bufsize))
   {
     ret = make_PyUnicode_from_ppd_string (self, buffer);
   } else {
-    ret = Py_None;
-    Py_INCREF (Py_None);
+    Py_RETURN_NONE;
   }
 
+  free (reason);
+  if (scheme)
+    free (scheme);
   free (buffer);
   return ret;
 }
@@ -296,20 +315,24 @@ PPD_localizeMarkerName (PPD *self, PyObject *args)
 {
 #if CUPS_VERSION_MAJOR > 1 || (CUPS_VERSION_MAJOR == 1 && CUPS_VERSION_MINOR >= 4)
   PyObject *ret;
-  const char *name;
+  PyObject *nameobj;
+  char *name;
   const char *lname;
 
-  if (!PyArg_ParseTuple (args, "s", &name))
+  if (!PyArg_ParseTuple (args, "O", &nameobj))
+    return NULL;
+
+  if (UTF8_from_PyObj (&name, nameobj) == NULL)
     return NULL;
 
   lname = ppdLocalizeMarkerName (self->ppd, name);
+  free (name);
 
   if (lname != NULL)
   {
     ret = make_PyUnicode_from_ppd_string (self, lname);
   } else {
-    ret = Py_None;
-    Py_INCREF (Py_None);
+    Py_RETURN_NONE;
   }
 
   return ret;
@@ -359,20 +382,29 @@ PPD_markOption (PPD *self, PyObject *args)
 static PyObject *
 PPD_conflicts (PPD *self)
 {
+#if PY_MAJOR_VERSION >= 3
+  return PyLong_FromLong (ppdConflicts (self->ppd));
+#else
   return PyInt_FromLong (ppdConflicts (self->ppd));
+#endif
 }
 
 static PyObject *
 PPD_findOption (PPD *self, PyObject *args)
 {
   PyObject *ret;
-  const char *option;
+  PyObject *optionobj;
+  char *option;
   ppd_option_t *opt;
 
-  if (!PyArg_ParseTuple (args, "s", &option))
+  if (!PyArg_ParseTuple (args, "O", &optionobj))
+    return NULL;
+
+  if (UTF8_from_PyObj (&option, optionobj) == NULL)
     return NULL;
 
   opt = ppdFindOption (self->ppd, option);
+  free (option);
   if (opt) {
     PyObject *args = Py_BuildValue ("()");
     PyObject *kwlist = Py_BuildValue ("{}");
@@ -385,8 +417,7 @@ PPD_findOption (PPD *self, PyObject *args)
     Py_INCREF (self);
     ret = (PyObject *) optobj;
   } else {
-    ret = Py_None;
-    Py_INCREF (ret);
+    Py_RETURN_NONE;
   }
 
   return ret;
@@ -396,16 +427,30 @@ static PyObject *
 PPD_findAttr (PPD *self, PyObject *args, PyObject *kwds)
 {
   PyObject *ret;
-  const char *name;
-  const char *spec = NULL;
+  PyObject *nameobj;
+  PyObject *specobj = NULL;
+  char *name;
+  char *spec = NULL;
   ppd_attr_t *attr;
   static char *kwlist[] = { "name", "spec", NULL };
 
-  if (!PyArg_ParseTupleAndKeywords (args, kwds, "s|s", kwlist,
-				    &name, &spec))
+  if (!PyArg_ParseTupleAndKeywords (args, kwds, "O|O", kwlist,
+				    &nameobj, &specobj))
     return NULL;
 
+  if (UTF8_from_PyObj (&name, nameobj) == NULL)
+    return NULL;
+
+  if (specobj)
+    if (UTF8_from_PyObj (&spec, specobj) == NULL) {
+      free (name);
+      return NULL;
+    }
+
   attr = ppdFindAttr (self->ppd, name, spec);
+  free (name);
+  if (spec)
+    free (spec);
   if (attr) {
     PyObject *largs = Py_BuildValue ("()");
     PyObject *lkwlist = Py_BuildValue ("{}");
@@ -418,8 +463,7 @@ PPD_findAttr (PPD *self, PyObject *args, PyObject *kwds)
     Py_INCREF (self);
     ret = (PyObject *) attrobj;
   } else {
-    ret = Py_None;
-    Py_INCREF (ret);
+    Py_RETURN_NONE;
   }
 
   return ret;
@@ -429,16 +473,30 @@ static PyObject *
 PPD_findNextAttr (PPD *self, PyObject *args, PyObject *kwds)
 {
   PyObject *ret;
-  const char *name;
-  const char *spec = NULL;
+  PyObject *nameobj;
+  PyObject *specobj = NULL;
+  char *name;
+  char *spec = NULL;
   ppd_attr_t *attr;
   static char *kwlist[] = { "name", "spec", NULL };
 
-  if (!PyArg_ParseTupleAndKeywords (args, kwds, "s|s", kwlist,
-				    &name, &spec))
+  if (!PyArg_ParseTupleAndKeywords (args, kwds, "O|O", kwlist,
+				    &nameobj, &specobj))
     return NULL;
 
+  if (UTF8_from_PyObj (&name, nameobj) == NULL)
+    return NULL;
+
+  if (specobj)
+    if (UTF8_from_PyObj (&spec, specobj) == NULL) {
+      free (name);
+      return NULL;
+    }
+
   attr = ppdFindNextAttr (self->ppd, name, spec);
+  free (name);
+  if (spec)
+    free (spec);
   if (attr) {
     PyObject *largs = Py_BuildValue ("()");
     PyObject *lkwlist = Py_BuildValue ("{}");
@@ -451,8 +509,7 @@ PPD_findNextAttr (PPD *self, PyObject *args, PyObject *kwds)
     Py_INCREF (self);
     ret = (PyObject *) attrobj;
   } else {
-    ret = Py_None;
-    Py_INCREF (ret);
+    Py_RETURN_NONE;
   }
 
   return ret;
@@ -550,11 +607,10 @@ PPD_emitString (PPD *self, PyObject *args)
   emitted = ppdEmitString(self->ppd, section, min_order);
 
   if (emitted) {
-    ret = PyString_FromString(emitted);
+    ret = PyUnicode_FromString (emitted);
     free (emitted);
   } else {
-    Py_INCREF (Py_None);
-    ret = Py_None;
+    Py_RETURN_NONE;
   }
 
   return ret;
@@ -574,9 +630,14 @@ PPD_emit (PPD *self, PyObject *args)
   if (!PyArg_ParseTuple (args, "Oi", &pyFile, &section))
     return NULL;
 
+#if PY_MAJOR_VERSION >= 3
+  int fd = PyObject_AsFileDescriptor(pyFile);
+  f = fdopen(fd, "w");
+#else
   f = PyFile_AsFile(pyFile);
+#endif
   if (!f)
-    return NULL;
+    return PyErr_SetFromErrno (PyExc_RuntimeError);
 
   if (!ppdEmit(self->ppd, f, section))
     Py_RETURN_NONE;
@@ -599,9 +660,14 @@ PPD_emitAfterOrder (PPD *self, PyObject *args)
   if (!PyArg_ParseTuple (args, "Oiif", &pyFile, &section, &limit, &min_order))
     return NULL;
 
+#if PY_MAJOR_VERSION >= 3
+  int fd = PyObject_AsFileDescriptor(pyFile);
+  f = fdopen(fd, "w");
+#else
   f = PyFile_AsFile(pyFile);
+#endif
   if (!f)
-    return NULL;
+    return PyErr_SetFromErrno (PyExc_RuntimeError);
 
   if (!ppdEmitAfterOrder(self->ppd, f, section, limit, min_order))
     Py_RETURN_NONE;
@@ -635,19 +701,37 @@ PPD_emitJCL (PPD *self, PyObject *args)
 {
   PyObject *pyFile;
   int job_id;
-  const char *user;
-  const char *title;
+  PyObject *userobj;
+  PyObject *titleobj;
+  char *user;
+  char *title;
   FILE *f;
 
-  if (!PyArg_ParseTuple (args, "Oiss", &pyFile, &job_id, &user, &title))
+  if (!PyArg_ParseTuple (args, "OiOO", &pyFile, &job_id, &userobj, &titleobj))
     return NULL;
 
-  f = PyFile_AsFile(pyFile);
-  if (!f)
+  if (UTF8_from_PyObj (&user, userobj) == NULL)
     return NULL;
 
+  if (UTF8_from_PyObj (&title, titleobj) == NULL) {
+    free (user);
+    return NULL;
+  }
+
+#if PY_MAJOR_VERSION >= 3
+  int fd = PyObject_AsFileDescriptor(pyFile);
+  f = fdopen(fd, "w");
+#else
+  f = PyFile_AsFile(pyFile);
+#endif
+  if (!f)
+    return PyErr_SetFromErrno (PyExc_RuntimeError);
+
   if (!ppdEmitJCL(self->ppd, f, job_id, user, title))
     Py_RETURN_NONE;
+
+  free (user);
+  free (title);
   return PyErr_SetFromErrno (PyExc_RuntimeError);
 }
 
@@ -664,9 +748,14 @@ PPD_emitJCLEnd (PPD *self, PyObject *args)
   if (!PyArg_ParseTuple (args, "O", &pyFile))
     return NULL;
 
+#if PY_MAJOR_VERSION >= 3
+  int fd = PyObject_AsFileDescriptor(pyFile);
+  f = fdopen(fd, "w");
+#else
   f = PyFile_AsFile(pyFile);
+#endif
   if (!f)
-    return NULL;
+    return PyErr_SetFromErrno (PyExc_RuntimeError);
 
   if (!ppdEmitJCLEnd(self->ppd, f))
     Py_RETURN_NONE;
@@ -685,16 +774,12 @@ PPD_writeFd (PPD *self, PyObject *args)
     return NULL;
 
   dfd = dup (fd);
-  if (dfd == -1) {
-    PyErr_SetFromErrno (PyExc_RuntimeError);
-    return NULL;
-  }
+  if (dfd == -1)
+    return PyErr_SetFromErrno (PyExc_RuntimeError);
 
   out = fdopen (dfd, "w");
-  if (!out) {
-    PyErr_SetFromErrno (PyExc_RuntimeError);
-    return NULL;
-  }
+  if (!out)
+    return PyErr_SetFromErrno (PyExc_RuntimeError);
 
   rewind (self->file);
   while (!feof (self->file)) {
@@ -978,8 +1063,7 @@ PyMethodDef PPD_methods[] =
 
 PyTypeObject cups_PPDType =
   {
-    PyObject_HEAD_INIT(NULL)
-    0,                         /*ob_size*/
+    PyVarObject_HEAD_INIT(NULL, 0)
     "cups.PPD",                /*tp_name*/
     sizeof(PPD),               /*tp_basicsize*/
     0,                         /*tp_itemsize*/
@@ -1051,7 +1135,7 @@ static void
 Option_dealloc (Option *self)
 {
   Py_XDECREF (self->ppd);
-  self->ob_type->tp_free ((PyObject *) self);
+  ((PyObject *)self)->ob_type->tp_free ((PyObject *) self);
 }
 
 static PyObject *
@@ -1059,10 +1143,16 @@ Option_repr (Option *self)
 {
   ppd_option_t *option = self->option;
   if (!option)
-    return PyString_FromString ("<cups.Option>");
+    return PyUnicode_FromString ("<cups.Option>");
 
-  return PyString_FromFormat ("<cups.Option %s=%s>",
-			      option->keyword, option->defchoice);
+  char buffer[256];
+  snprintf (buffer, 256, "<cups.Option %s=%s>",
+			  option->keyword, option->defchoice);
+#if PY_MAJOR_VERSION >= 3
+    return PyUnicode_FromString (buffer);
+#else
+    return PyBytes_FromString (buffer);
+#endif
 }
 
 ////////////
@@ -1115,7 +1205,11 @@ Option_getUI (Option *self, void *closure)
     Py_RETURN_NONE;
   }
 
+#if PY_MAJOR_VERSION >= 3
+  return PyLong_FromLong (self->option->ui);
+#else
   return PyInt_FromLong (self->option->ui);
+#endif
 }
 
 static PyObject *
@@ -1209,8 +1303,7 @@ PyMethodDef Option_methods[] =
 
 PyTypeObject cups_OptionType =
   {
-    PyObject_HEAD_INIT(NULL)
-    0,                         /*ob_size*/
+    PyVarObject_HEAD_INIT(NULL, 0)
     "cups.Option",             /*tp_name*/
     sizeof(Option),            /*tp_basicsize*/
     0,                         /*tp_itemsize*/
@@ -1289,7 +1382,7 @@ static void
 Group_dealloc (Group *self)
 {
   Py_XDECREF (self->ppd);
-  self->ob_type->tp_free ((PyObject *) self);
+  ((PyObject *)self)->ob_type->tp_free ((PyObject *) self);
 }
 
 static PyObject *
@@ -1297,9 +1390,15 @@ Group_repr (Group *self)
 {
   ppd_group_t *group = self->group;
   if (!group)
-    return PyString_FromString ("<cups.Group>");
+    return PyUnicode_FromString ("<cups.Group>");
 
-  return PyString_FromFormat ("<cups.Group %s>", group->name);
+  char buffer[256];
+  snprintf (buffer, 256, "<cups.Group %s>", group->name);
+#if PY_MAJOR_VERSION >= 3
+    return PyUnicode_FromString (buffer);
+#else
+    return PyBytes_FromString (buffer);
+#endif
 }
 
 ///////////
@@ -1410,8 +1509,7 @@ PyMethodDef Group_methods[] =
 
 PyTypeObject cups_GroupType =
   {
-    PyObject_HEAD_INIT(NULL)
-    0,                         /*ob_size*/
+    PyVarObject_HEAD_INIT(NULL, 0)
     "cups.Group",              /*tp_name*/
     sizeof(Group),             /*tp_basicsize*/
     0,                         /*tp_itemsize*/
@@ -1485,7 +1583,7 @@ static void
 Constraint_dealloc (Constraint *self)
 {
   Py_XDECREF (self->ppd);
-  self->ob_type->tp_free ((PyObject *) self);
+  ((PyObject *)self)->ob_type->tp_free ((PyObject *) self);
 }
 
 ////////////////
@@ -1555,8 +1653,7 @@ PyGetSetDef Constraint_getseters[] =
 
 PyTypeObject cups_ConstraintType =
   {
-    PyObject_HEAD_INIT(NULL)
-    0,                         /*ob_size*/
+    PyVarObject_HEAD_INIT(NULL, 0)
     "cups.Constraint",         /*tp_name*/
     sizeof(Constraint),        /*tp_basicsize*/
     0,                         /*tp_itemsize*/
@@ -1630,7 +1727,7 @@ static void
 Attribute_dealloc (Attribute *self)
 {
   Py_XDECREF (self->ppd);
-  self->ob_type->tp_free ((PyObject *) self);
+  ((PyObject *)self)->ob_type->tp_free ((PyObject *) self);
 }
 
 static PyObject *
@@ -1638,12 +1735,18 @@ Attribute_repr (Attribute *self)
 {
   ppd_attr_t *attribute = self->attribute;
   if (!attribute)
-    return PyString_FromString ("<cups.Attribute>");
+    return PyUnicode_FromString ("<cups.Attribute>");
 
-  return PyString_FromFormat ("<cups.Attribute *%s%s%s>",
-			      attribute->name,
-			      attribute->spec[0] != '\0' ? " ": "",
-			      attribute->spec);
+  char buffer[256];
+  snprintf (buffer, 256, "<cups.Attribute *%s%s%s>",
+			  attribute->name,
+			  attribute->spec[0] != '\0' ? " ": "",
+			  attribute->spec);
+#if PY_MAJOR_VERSION >= 3
+    return PyUnicode_FromString (buffer);
+#else
+    return PyBytes_FromString (buffer);
+#endif
 }
 
 ///////////////
@@ -1713,8 +1816,7 @@ PyGetSetDef Attribute_getseters[] =
 
 PyTypeObject cups_AttributeType =
   {
-    PyObject_HEAD_INIT(NULL)
-    0,                         /*ob_size*/
+    PyVarObject_HEAD_INIT(NULL, 0)
     "cups.Attribute",          /*tp_name*/
     sizeof(Attribute),         /*tp_basicsize*/
     0,                         /*tp_itemsize*/
